{% extends "layout.html" %}

{% block body %}
	<div id="container-not-an-ad">
		<div id="ad">
			<div class="row">
				<div class="col-xs-2">
					<img id="adlogo" src="/static/images/logo.png" alt="Lab Logo" />
				</div>
				<div class="col-xs-10">
					<h1>Returning to HIT window.</h1>
					<p>You are being redirected back to the HIT window, which should refresh.
					If you do not see the HIT submission form when you return, try refreshing
					the page.</p>
				</div>
			</div>
		</div>
	</div>
{% endblock %}

{% block scripts %}
	<script>
		window.opener.location.reload(true);
		window.close();
	</script>
{% endblock %}
